@@ -1,27 +1,35 @@
-//#define DL_DEBUG
-/* Run an ELF binary on a linux system.
+/* vi: set sw=4 ts=4: */
+/* i386 ELF shared library loader suppport
+ *
+ * Copyright (C) 2001-2002,  David A. Schleef
+ *
+ * All rights reserved.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer.
+ * 2. The name of the above contributors may not be
+ *    used to endorse or promote products derived from this software
+ *    without specific prior written permission.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE CONTRIBUTORS ``AS IS'' AND
+ * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
+ * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
+ * ARE DISCLAIMED.  IN NO EVENT SHALL THE CONTRIBUTORS BE LIABLE
+ * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
+ * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
+ * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
+ * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
+ * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
+ * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
+ * SUCH DAMAGE.
+ */
 
-   Copyright (C) 1993, Eric Youngdale.
-
-   This program is free software; you can redistribute it and/or modify
-   it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2, or (at your option)
-   any later version.
-
-   This program is distributed in the hope that it will be useful,
-   but WITHOUT ANY WARRANTY; without even the implied warranty of
-   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-   GNU General Public License for more details.
-
-   You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
-
-#ifndef VERBOSE_DLINKER
-#define VERBOSE_DLINKER
-#endif
-#ifdef VERBOSE_DLINKER
-static char *_dl_reltypes[] =
+/* FIXME -- Disable this when __SUPPORT_LD_DEBUG__ is undefined */
+#if defined (__SUPPORT_LD_DEBUG__)
+static const char *_dl_reltypes[] =
 	{ "R_PPC_NONE", "R_PPC_ADDR32", "R_PPC_ADDR24", "R_PPC_ADDR16",
 	"R_PPC_ADDR16_LO", "R_PPC_ADDR16_HI", "R_PPC_ADDR16_HA",
 	"R_PPC_ADDR14", "R_PPC_ADDR14_BRTAKEN", "R_PPC_ADDR14_BRNTAKEN",
@@ -49,10 +57,10 @@ static char *_dl_reltypes[] =
    working. */
 
 
-#ifdef DL_DEBUG_SYMBOLS
+#ifdef __SUPPORT_LD_DEBUG__
 static void debug_sym(Elf32_Sym *symtab,char *strtab,int symtab_index);
 static void debug_reloc(ELF_RELOC *rpnt);
-#define DPRINTF(fmt,args...) _dl_dprintf(2,fmt,args)
+#define DPRINTF(fmt,args...) if (_dl_debug) _dl_dprintf(2,fmt,args)
 #else
 #define debug_sym(a,b,c)
 #define debug_reloc(a)
@@ -144,9 +152,13 @@ unsigned long _dl_linux_resolver(struct elf_resolve *tpnt, int reloc_entry)
 	//debug_reloc(this_reloc);
 
 	if (reloc_type != R_PPC_JMP_SLOT) {
-		_dl_dprintf(2, "%s: Incorrect relocation type [%s] in jump relocations\n",
+#if defined (__SUPPORT_LD_DEBUG__)
+		_dl_dprintf(2, "%s: Incorrect relocation type [%s] in jump relocation\n",
 			_dl_progname,
 			(reloc_type<N_RELTYPES)?_dl_reltypes[reloc_type]:"unknown");
+#else
+		_dl_dprintf(2, "%s: Incorrect relocation type in jump relocation\n", _dl_progname);
+#endif
 		_dl_exit(1);
 	};
 
@@ -161,7 +173,7 @@ unsigned long _dl_linux_resolver(struct elf_resolve *tpnt, int reloc_entry)
 	/* Get the address of the GOT entry */
 	targ_addr = (unsigned long) _dl_find_hash(
 		strtab + symtab[symtab_index].st_name, 
-		tpnt->symbol_scope, tpnt, 0);
+		tpnt->symbol_scope, tpnt, resolver);
 	if (!targ_addr) {
 		_dl_dprintf(2, "%s: can't resolve symbol '%s'\n", 
 			_dl_progname, strtab + symtab[symtab_index].st_name);
@@ -276,7 +288,7 @@ void _dl_parse_lazy_relocation_information(struct elf_resolve *tpnt,
 		default:
 			_dl_dprintf(2, "%s: (LAZY) can't handle reloc type ", 
 				_dl_progname);
-#ifdef VERBOSE_DLINKER
+#if defined (__SUPPORT_LD_DEBUG__)
 			_dl_dprintf(2, "%s ", _dl_reltypes[reloc_type]);
 #endif
 			if (symtab_index)
@@ -338,7 +350,7 @@ int _dl_parse_relocation_information(struct elf_resolve *tpnt,
 
 			symbol_addr = (unsigned long) _dl_find_hash(strtab + symtab[symtab_index].st_name, 
 					tpnt->symbol_scope,
-					(reloc_type == R_PPC_JMP_SLOT ? tpnt : NULL), 0);
+					(reloc_type == R_PPC_JMP_SLOT ? tpnt : NULL), symbolrel);
 
 			/*
 			 * We want to allow undefined references to weak symbols - this might
@@ -424,7 +436,7 @@ int _dl_parse_relocation_information(struct elf_resolve *tpnt,
 			break;
 		default:
 			_dl_dprintf(2, "%s: can't handle reloc type ", _dl_progname);
-#ifdef VERBOSE_DLINKER
+#if defined (__SUPPORT_LD_DEBUG__)
 			_dl_dprintf(2, "%s ", _dl_reltypes[reloc_type]);
 #endif
 			if (symtab_index)
@@ -499,7 +511,7 @@ int _dl_parse_copy_information(struct dyn_elf *xpnt, unsigned long rel_addr,
 
 			symbol_addr = (unsigned long) _dl_find_hash(strtab + 
 				symtab[symtab_index].st_name, xpnt->next, 
-				NULL, 1);
+				NULL, copyrel);
 			if (!symbol_addr) {
 				_dl_dprintf(2, "%s: can't resolve symbol '%s'\n", 
 					_dl_progname, strtab + symtab[symtab_index].st_name);
@@ -551,29 +563,33 @@ static void fixup_jmpslot(unsigned long reloc_addr, unsigned long targ_addr)
 #endif
 
 
-#ifdef DL_DEBUG_SYMBOLS
+#ifdef __SUPPORT_LD_DEBUG__
 static void debug_sym(Elf32_Sym *symtab,char *strtab,int symtab_index)
 {
-	if(symtab_index){
-		_dl_dprintf(2, "sym: name=%s value=%x size=%x info=%x other=%x shndx=%x\n",
-			strtab + symtab[symtab_index].st_name,
-			symtab[symtab_index].st_value,
-			symtab[symtab_index].st_size,
-			symtab[symtab_index].st_info,
-			symtab[symtab_index].st_other,
-			symtab[symtab_index].st_shndx);
-	}else{
-		_dl_dprintf(2, "sym: null\n");
+	if (_dl_debug_symbols) {
+		if(symtab_index){
+			_dl_dprintf(2, "sym: name=%s value=%x size=%x info=%x other=%x shndx=%x\n",
+					strtab + symtab[symtab_index].st_name,
+					symtab[symtab_index].st_value,
+					symtab[symtab_index].st_size,
+					symtab[symtab_index].st_info,
+					symtab[symtab_index].st_other,
+					symtab[symtab_index].st_shndx);
+		}else{
+			_dl_dprintf(2, "sym: null\n");
+		}
 	}
 }
 
 static void debug_reloc(ELF_RELOC *rpnt)
 {
-	_dl_dprintf(2, "reloc: offset=%x type=%x sym=%x addend=%x\n",
-		rpnt->r_offset,
-		ELF32_R_TYPE(rpnt->r_info),
-		ELF32_R_SYM(rpnt->r_info),
-		rpnt->r_addend);
+	if (_dl_debug_reloc) {
+		_dl_dprintf(2, "reloc: offset=%x type=%x sym=%x addend=%x\n",
+				rpnt->r_offset,
+				ELF32_R_TYPE(rpnt->r_info),
+				ELF32_R_SYM(rpnt->r_info),
+				rpnt->r_addend);
+	}
 }
 
 #endif
