@@ -31,15 +31,45 @@
  *
  *  ATTENTION!   ATTENTION!   ATTENTION!   ATTENTION!   ATTENTION! */
 
-/* 4-01-2002
+
+/* April 1, 2002
  * Initialize thread locks for fake files in vsnprintf and vdprintf.
  *    reported by Erik Andersen (andersen@codepoet.com)
  * Fix an arg promotion handling bug in _do_one_spec for %c. 
  *    reported by Ilguiz Latypov <ilatypov@superbt.com>
  *
- * 5-10-2002
+ * May 10, 2002
  * Remove __isdigit and use new ctype.h version.
  * Add conditional setting of QUAL_CHARS for size_t and ptrdiff_t.
+ *
+ * Aug 16, 2002
+ * Fix two problems that showed up with the python 2.2.1 tests; one
+ *    involving %o and one involving %f.
+ *
+ * Oct 28, 2002
+ * Fix a problem in vasprintf (reported by vodz a while back) when built
+ *    without custom stream support.  In that case, it is necessary to do
+ *    a va_copy.
+ * Make sure each va_copy has a matching va_end, as required by C99.
+ *
+ * Nov 4, 2002
+ * Add locale-specific grouping support for integer decimal conversion.
+ * Add locale-specific decimal point support for floating point conversion.
+ *   Note: grouping will have to wait for _dtostr() rewrite.
+ * Add printf wchar support for %lc (%C) and %ls (%S).
+ * Require printf format strings to be valid multibyte strings beginning and
+ *   ending in their initial shift state, as per the stds.
+ */
+
+/* TODO:
+ *
+ * Should we validate that *printf format strings are valid multibyte
+ *   strings in the current locale?  ANSI/ISO C99 seems to imply this
+ *   and Plauger's printf implementation in his Standard C Library book
+ *   treats this as an error.
+ *
+ * Implement %a, %A, and locale-specific grouping for the printf floating
+ *   point conversions.  To be done in the rewrite of _dtostr().
  */
 
 
@@ -61,9 +91,14 @@
 #include <printf.h>
 
 #ifdef __STDIO_THREADSAFE
+#include <stdio_ext.h>
 #include <pthread.h>
 #endif /* __STDIO_THREADSAFE */
 
+#ifdef __UCLIBC_HAS_WCHAR__
+#include <wchar.h>
+#endif /* __UCLIBC_HAS_WCHAR__ */
+
 /**********************************************************************/
 
 /* These provide some control over printf's feature set */
@@ -103,6 +138,26 @@
 #define MAX_ARGS		MAX_POS_ARGS
 #endif
 
+/**********************************************************************/
+/* Deal with pre-C99 compilers. */
+
+#ifndef va_copy
+
+#ifdef __va_copy
+#define va_copy(A,B)	__va_copy(A,B)
+#else
+	/* TODO -- maybe create a bits/vacopy.h for arch specific versions
+	 * to ensure we get the right behavior?  Either that or fall back
+	 * on the portable (but costly in size) method of using a va_list *.
+	 * That means a pointer derefs in the va_arg() invocations... */
+#warning neither va_copy or __va_copy is defined.  using a simple copy instead...
+	/* the glibc manual suggests that this will usually suffice when
+        __va_copy doesn't exist.  */
+#define va_copy(A,B)	A = B
+#endif
+
+#endif /* va_copy */
+
 /**********************************************************************/
 
 #define __PA_FLAG_INTMASK \
@@ -304,7 +359,7 @@ typedef struct {
 extern size_t _dtostr(FILE * fp, long double x, struct printf_info *info);
 #endif
 
-#define _outnstr(stream, string, len)	_stdio_fwrite(s, len, stream)	/* TODO */
+#define _outnstr(stream, string, len)	_stdio_fwrite(string, len, stream)	/* TODO */
 
 extern int _do_one_spec(FILE * __restrict stream, ppfs_t *ppfs, int *count);
 
@@ -387,8 +442,6 @@ size_t parse_printf_format(register const char *template,
  * In other words, we don't currently support glibc's 'I' flag.
  * We do accept it, but it is currently ignored. */
 
-/* TODO -- Implement grouping flag... on hold (locale support). */
-
 int vfprintf(FILE * __restrict stream, register const char * __restrict format,
 			 va_list arg)
 {
@@ -402,10 +455,10 @@ int vfprintf(FILE * __restrict stream, register const char * __restrict format,
 	s = format;
 
 	if (_ppfs_init(&ppfs, format) < 0) { /* Bad format string. */
-		_outnstr(stream, format, strlen(format));
+		_outnstr(stream, ppfs.fmtpos, strlen(ppfs.fmtpos));
 		count = -1;
 	} else {
-		_ppfs_prepargs(&ppfs, arg);
+		_ppfs_prepargs(&ppfs, arg);	/* This did a va_copy!!! */
 
 		do {
 			while (*format && (*format != '%')) {
@@ -438,6 +491,8 @@ int vfprintf(FILE * __restrict stream, register const char * __restrict format,
 				++format;
 			}
 		} while (1);
+
+		va_end(ppfs.arg);		/* Need to clean up after va_copy! */
 	}
 
 	__STDIO_THREADUNLOCK(stream);
@@ -450,11 +505,29 @@ int vfprintf(FILE * __restrict stream, register const char * __restrict format,
 
 int _ppfs_init(register ppfs_t *ppfs, const char *fmt0)
 {
+#if defined(__UCLIBC_HAS_WCHAR__) && defined(__UCLIBC_HAS_LOCALE__)
+	static const char invalid_mbs[] = "Invalid multibyte format string.";
+#endif /* defined(__UCLIBC_HAS_WCHAR__) && defined(__UCLIBC_HAS_LOCALE__) */
 	int r;
 
 	/* First, zero out everything... argnumber[], argtype[], argptr[] */
 	memset(ppfs, 0, sizeof(ppfs_t)); /* TODO: nonportable???? */
 	--ppfs->maxposarg;			/* set to -1 */
+	ppfs->fmtpos = fmt0;
+#if defined(__UCLIBC_HAS_WCHAR__) && defined(__UCLIBC_HAS_LOCALE__)
+	/* Note: We don't need to check if we don't have wide chars or we only
+	 * support the C locale. */
+	{
+		mbstate_t mbstate;
+		const char *p;
+		mbstate.mask = 0;	/* Initialize the mbstate. */
+		p = fmt0;
+		if (mbsrtowcs(NULL, &p, SIZE_MAX, &mbstate) == ((size_t)(-1))) {
+			ppfs->fmtpos = invalid_mbs;
+			return -1;
+		}
+	}
+#endif /* defined(__UCLIBC_HAS_WCHAR__) && defined(__UCLIBC_HAS_LOCALE__) */
 	/* now set all argtypes to no-arg */
 	{
 #if 1
@@ -518,18 +591,7 @@ void _ppfs_prepargs(register ppfs_t *ppfs, va_list arg)
 {
 	int i;
 
-#ifdef __va_copy
-	__va_copy(ppfs->arg, arg);
-#else
-	/* TODO -- maybe create a bits/vacopy.h for arch specific versions
-	 * to ensure we get the right behavior?  Either that or fall back
-	 * on the portable (but costly in size) method of using a va_list *.
-	 * That means a pointer derefs in the va_arg() invocations... */
-#warning __va_copy is not defined, using a simple copy instead...
-	/* the glibc manual suggests that this will usually suffice when
-        __va_copy doesn't exist.  */
-	ppfs->arg = arg;
-#endif
+	va_copy(ppfs->arg, arg);
 
 	if ((i = ppfs->maxposarg) > 0)  { /* init for positional args */
 		ppfs->num_data_args = i;
@@ -1078,13 +1140,25 @@ int _do_one_spec(FILE * __restrict stream, register ppfs_t *ppfs, int *count)
 	const void * argptr[MAX_ARGS_PER_SPEC];
 #endif
 	int *argtype;
+#ifdef __UCLIBC_HAS_WCHAR__
+	const wchar_t *ws = NULL;
+	mbstate_t mbstate;
+#endif /* __UCLIBC_HAS_WCHAR__ */
 	size_t slen;
 	int base;
 	int numpad;
+	int alphacase;
 	int numfill = 0;			/* TODO: fix */
 	int prefix_num = PREFIX_NONE;
 	char padchar = ' ';
-	char buf[64];
+#ifdef __UCLIBC_MJN3_ONLY__
+#warning REMINDER: buf size
+#endif
+	/* TODO: buf needs to be big enough for any possible error return strings
+	 * and also for any locale-grouped long long integer strings generated.
+	 * This should be large enough for any of the current archs/locales, but
+	 * eventually this should be handled robustly. */
+	char buf[128];
 
 #ifdef NDEBUG
 	_ppfs_parsespec(ppfs);
@@ -1125,8 +1199,16 @@ int _do_one_spec(FILE * __restrict stream, register ppfs_t *ppfs, int *count)
 			return 0;
 		}
 		if (ppfs->conv_num <= CONV_i) {	/* pointer or (un)signed int */
-			base = spec_base[(int)(ppfs->conv_num - CONV_p)];
+			alphacase = __UIM_LOWER;
+			if (((base = spec_base[(int)(ppfs->conv_num - CONV_p)]) == 10)
+				&& (PRINT_INFO_FLAG_VAL(&(ppfs->info),group))
+				) {
+				alphacase = __UIM_GROUP;
+			}
 			if (ppfs->conv_num <= CONV_u) { /* pointer or unsigned int */
+				if (ppfs->conv_num == CONV_X) {
+					alphacase = __UIM_UPPER;
+				}
 				if (ppfs->conv_num == CONV_p) { /* pointer */
 					prefix_num = PREFIX_LWR_X;
 				} else {		/* unsigned int */
@@ -1140,9 +1222,7 @@ int _do_one_spec(FILE * __restrict stream, register ppfs_t *ppfs, int *count)
 			s = _uintmaxtostr(buf + sizeof(buf) - 1,
 							  (uintmax_t)
 							  _load_inttype(*argtype & __PA_INTMASK,
-											*argptr, base), base,
-							  ((ppfs->conv_num == CONV_X)
-							   ? __UIM_UPPER : __UIM_LOWER));
+											*argptr, base), base, alphacase);
 			if (ppfs->conv_num > CONV_u) { /* signed int */
 				if (*s == '-') {
 					PRINT_INFO_SET_FLAG(&(ppfs->info),showsign);
@@ -1163,7 +1243,7 @@ int _do_one_spec(FILE * __restrict stream, register ppfs_t *ppfs, int *count)
 				if (ppfs->conv_num == CONV_X) {
 					prefix_num = PREFIX_UPR_X;
 				}
-				if ((ppfs->conv_num == CONV_o) && (numfill < slen)) {
+				if ((ppfs->conv_num == CONV_o) && (numfill <= slen)) {
 					numfill = ((*s == '0') ? 1 : slen + 1);
 				}
 			}
@@ -1189,18 +1269,39 @@ int _do_one_spec(FILE * __restrict stream, register ppfs_t *ppfs, int *count)
 							  &ppfs->info);
 			return 0;
 #else  /* __STDIO_PRINTF_FLOAT */
-			return -1;			/* TODO -- try ton continue? */
+			return -1;			/* TODO -- try to continue? */
 #endif /* __STDIO_PRINTF_FLOAT */
 		} else if (ppfs->conv_num <= CONV_S) {	/* wide char or string */
-#if 1
-			return -1;			/* TODO -- wide */
-#else
+#ifdef __UCLIBC_HAS_WCHAR__
+			mbstate.mask = 0;	/* Initialize the mbstate. */
 			if (ppfs->conv_num == CONV_S) { /* wide string */
-
+				if (!(ws = *((const wchar_t **) *argptr))) {
+					goto NULL_STRING;
+				}
+				/* We use an awful uClibc-specific hack here, passing
+				 * (char*) &ws as the conversion destination.  This signals
+				 * uClibc's wcsrtombs that we want a "restricted" length
+				 * such that the mbs fits in a buffer of the specified
+				 * size with no partial conversions. */
+				if ((slen = wcsrtombs((char *) &ws, &ws, /* Use awful hack! */
+									  ((ppfs->info.prec >= 0)
+									   ? ppfs->info.prec
+									   : SIZE_MAX), &mbstate))
+					== ((size_t)-1)
+					) {
+					return -1;	/* EILSEQ */
+				}
 			} else {			/* wide char */
-				
+				s = buf;
+				slen = wcrtomb(s, (*((const wchar_t *) *argptr)), &mbstate);
+				if (slen == ((size_t)-1)) {
+					return -1;	/* EILSEQ */
+				}
+				s[slen] = 0;	/* TODO - Is this necessary? */
 			}
-#endif
+#else  /* __UCLIBC_HAS_WCHAR__ */
+			return -1;
+#endif /* __UCLIBC_HAS_WCHAR__ */
 		} else if (ppfs->conv_num <= CONV_s) {	/* char or string */
 			if (ppfs->conv_num == CONV_s) { /* string */
 				s = *((char **) (*argptr));
@@ -1209,11 +1310,12 @@ int _do_one_spec(FILE * __restrict stream, register ppfs_t *ppfs, int *count)
 					slen = strnlen(s, ((ppfs->info.prec >= 0)
 									   ? ppfs->info.prec : SIZE_MAX));
 				} else {
+				NULL_STRING:
 					s = "(null)";
 					slen = 6;
 				}
 			} else {			/* char */
-				s = (char *) buf;
+				s = buf;
 				*s = (unsigned char)(*((const int *) *argptr));
 				s[1] = 0;
 				slen = 1;
@@ -1267,7 +1369,24 @@ int _do_one_spec(FILE * __restrict stream, register ppfs_t *ppfs, int *count)
 		}
 		output(stream, prefix + prefix_num);
 		_charpad(stream, '0', numfill);
+#ifdef __UCLIBC_HAS_WCHAR__
+		if (!ws) {
+			_outnstr(stream, s, slen);
+		} else {				/* wide string */
+			size_t t;
+			mbstate.mask = 0;	/* Initialize the mbstate. */
+			while (slen) {
+				t = (slen <= sizeof(buf)) ? slen : sizeof(buf);
+				t = wcsrtombs(buf, &ws, t, &mbstate);
+				assert (t != ((size_t)(-1)));
+				_outnstr(stream, buf, t);
+				slen -= t;
+			}
+			ws = NULL;			/* Reset ws. */
+		}
+#else  /* __UCLIBC_HAS_WCHAR__ */
 		_outnstr(stream, s, slen);
+#endif /* __UCLIBC_HAS_WCHAR__ */
 		_charpad(stream, ' ', numpad);
 	}
 
@@ -1313,6 +1432,7 @@ int vsnprintf(char *__restrict buf, size_t size,
 #endif /* __STDIO_MBSTATE */
 
 #ifdef __STDIO_THREADSAFE
+	f.user_locking = 0;
 	__stdio_init_mutex(&f.lock);
 #endif
 
@@ -1392,6 +1512,7 @@ int vsnprintf(char *__restrict buf, size_t size,
 #endif /* __STDIO_MBSTATE */
 
 #ifdef __STDIO_THREADSAFE
+	f.user_locking = 0;
 	__stdio_init_mutex(&f.lock);
 #endif
 
@@ -1439,6 +1560,7 @@ int vdprintf(int filedes, const char * __restrict format, va_list arg)
 #endif /* __STDIO_MBSTATE */
 
 #ifdef __STDIO_THREADSAFE
+	f.user_locking = 0;
 	__stdio_init_mutex(&f.lock);
 #endif
 
@@ -1464,8 +1586,12 @@ int vasprintf(char **__restrict buf, const char * __restrict format,
 	 * only does one malloc.  This can be a problem though when custom printf
 	 * specs or the %m specifier are involved because the results of the
 	 * second call might be different from the first. */
+	va_list arg2;
+	int rv;
 
- 	int rv = vsnprintf(NULL, 0, format, arg);
+	va_copy(arg2, arg);
+ 	rv = vsnprintf(NULL, 0, format, arg2);
+	va_end(arg2);
 
 	return (((rv >= 0) && ((*buf = malloc(++rv)) != NULL))
 			? vsnprintf(*buf, rv, format, arg)
@@ -1749,6 +1875,15 @@ static const char *fmts[] = {
 };
 
 /*****************************************************************************/
+#include <locale.h>
+
+#ifdef __UCLIBC_MJN3_ONLY__
+#warning REMINDER: implement grouping for floating point
+#endif
+
+#ifndef __LOCALE_C_ONLY
+#define CUR_LOCALE			(__global_locale)
+#endif /* __LOCALE_C_ONLY */
 
 size_t _dtostr(FILE * fp, long double x, struct printf_info *info)
 {
@@ -1892,6 +2027,11 @@ size_t _dtostr(FILE * fp, long double x, struct printf_info *info)
 
 	if (mode == 'f') {
 		round += exp;
+		if (round < -1) {
+			memset(buf, '0', MAX_DIGITS);
+		    exp = -1;
+		    round = -1;
+		}
 	}
 
 	s = buf;
@@ -1966,9 +2106,15 @@ size_t _dtostr(FILE * fp, long double x, struct printf_info *info)
 	if (PRINT_INFO_FLAG_VAL(info,alt)
 /*  		flag[FLAG_HASH] */
 		|| (i) || ((o_mode != 'g') && (preci > 0))) {
+#ifdef __LOCALE_C_ONLY
 		*pdrvr++ = 2;			/* need decimal */
 		*ppc++ = 1;				/* needed for width calc */
 		ppc++;
+#else  /* __LOCALE_C_ONLY */
+		*pdrvr++ = 1;
+		*ppc++ = strlen(CUR_LOCALE.decimal_point);
+		*ppc++ = (INT_OR_PTR)(CUR_LOCALE.decimal_point);
+#endif /* __LOCALE_C_ONLY */
 	}
 
 	if (++o_exp < 0) {			/* have 0s right of decimal */
